@@ -11,12 +11,15 @@ from __future__ import annotations  # Make typing work for the enclosing class
 __author__ = "Jonas Van Der Donckt, Emiel Deprost, Jeroen Van Der Donckt"
 
 import os
-import dill
-import pandas as pd
-import numpy as np
-
+import uuid
+from copy import deepcopy
 from pathlib import Path
 from typing import Dict, List, Optional, Tuple, Union
+
+import dill
+import traceback
+import numpy as np
+import pandas as pd
 from pathos.multiprocessing import ProcessPool
 from tqdm.auto import tqdm
 
@@ -25,8 +28,8 @@ from .logger import logger
 from .strided_rolling import StridedRolling
 from ..features.function_wrapper import FuncWrapper
 from ..utils.data import to_list, to_series_list, flatten
-from ..utils.time import timedelta_to_str
 from ..utils.logging import delete_logging_handlers, add_logging_handler
+from ..utils.time import timedelta_to_str
 
 
 class FeatureCollection:
@@ -37,6 +40,23 @@ class FeatureCollection:
     feature_descriptors : Union[FeatureDescriptor, MultipleFeatureDescriptors, List[Union[FeatureDescriptor, MultipleFeatureDescriptors]]], optional
         Initial (list of) feature(s) to add to collection, by default None
 
+
+    Notes
+    -----
+    * The `series_name` property of the `FeatureDescriptor`s should **not withhold a "|"
+      character**, since "|" is used to join the series names of features which use
+      multiple series as input).<br>
+      e.g.<br>
+        * `ACC|x` is **not** allowed as series name, as this is ambiguous and could
+          represent that this feature is constructed with a combination of the `ACC`
+          and `x` signal.<br>
+          Note that `max|feat` is allowed as feature output name.
+    * Both the `series_name` and `output_name` property of the `FeatureDescriptor`s
+      **should not withhold "__"** in its string representations. This constraint is
+      mainly made for readability purposes.
+
+    The two statements above will be asserted
+
     """
 
     def __init__(
@@ -92,9 +112,17 @@ class FeatureCollection:
             The feature that will be added to this feature collection.
 
         """
+        # Check whether the `|` is not present in the series
+        assert not any('|' in s_name for s_name in feature.get_required_series())
+        # Check whether the '__" is not present in the series and function output names
+        assert not any(
+            '__' in output_name for output_name in feature.function.output_names
+        )
+        assert not any('__' in s_name for s_name in feature.get_required_series())
+
         series_win_stride_key = self._get_collection_key(feature)
         if series_win_stride_key in self._feature_desc_dict.keys():
-            added_output_names = flatten(                
+            added_output_names = flatten(
                 f.function.output_names
                 for f in self._feature_desc_dict[series_win_stride_key]
             )
@@ -111,7 +139,7 @@ class FeatureCollection:
     def add(
         self,
         features: Union[
-            FeatureDescriptor, 
+            FeatureDescriptor,
             MultipleFeatureDescriptors,
             FeatureCollection,
             List[
@@ -286,15 +314,19 @@ class FeatureCollection:
         )
         nb_stroll_funcs = self._get_stroll_feat_length()
 
-        if n_jobs is None: 
+        if n_jobs is None:
             n_jobs = os.cpu_count()
         n_jobs = min(n_jobs, nb_stroll_funcs)
 
+        calculated_feature_list = None
         if n_jobs in [0, 1]:
             idxs = range(nb_stroll_funcs)
             if show_progress:
                 idxs = tqdm(idxs)
-            calculated_feature_list = [self._executor(idx) for idx in idxs]
+            try:
+                calculated_feature_list = [self._executor(idx) for idx in idxs]
+            except:
+                traceback.print_exc()
         else:
             # https://pathos.readthedocs.io/en/latest/pathos.html#usage
             with ProcessPool(nodes=n_jobs, source=True) as pool:
@@ -304,12 +336,22 @@ class FeatureCollection:
                 )
                 if show_progress:
                     results = tqdm(results, total=self._get_stroll_feat_length())
-                calculated_feature_list = [f for f in results]
-                # Close & join - see: https://github.com/uqfoundation/pathos/issues/131
-                pool.close()
-                pool.join()
-                # Clear because: https://github.com/uqfoundation/pathos/issues/111
-                pool.clear()
+                try:
+                    calculated_feature_list = [f for f in results]
+                except:
+                    traceback.print_exc()
+                finally:
+                    # Close & join - see: https://github.com/uqfoundation/pathos/issues/131
+                    pool.close()
+                    pool.join()
+                    # Clear because: https://github.com/uqfoundation/pathos/issues/111
+                    pool.clear()
+
+        if calculated_feature_list is None:
+            raise RuntimeError(
+                "Feature Extraction halted due to error while extracting one (or multiple) feature(s)! " +
+                "See stack trace above."
+                )
 
         if return_df:
             return pd.concat(calculated_feature_list, axis=1, join="outer", copy=False)
@@ -334,6 +376,74 @@ class FeatureCollection:
         with open(file_path, "wb") as f:
             dill.dump(self, f, recurse=True)
 
+    def reduce(self, feat_cols_to_keep: List[str]) -> FeatureCollection:
+        """Create a reduced FeatureCollection instance based on feat_cols_to_keep.
+
+        For example, this is useful to optimize feature-extraction inference
+        (for your selected features) after performing a feature-selection procedure.
+
+        Parameters
+        ----------
+        feat_cols_to_keep: List[str]
+            A subset of the feature collection instance its column names.
+            This corresponds to the columns / names of the output from `calculate`
+            method that you want to keep.
+
+        Returns
+        -------
+        FeatureCollection
+            A new FeatureCollection object, which only withholds the FeatureDescriptors
+            which constitute the `feat_cols_to_keep` output.
+
+        Note
+        ----
+        Some FeatureDescriptor objects may have multiple **output-names**.<br>
+        Hence, if you only want to retain _a subset_ of that FeatureDescriptor its
+        feature outputs, you will still get **all features** as the new
+        FeatureCollection is constructed by applying a filter on de FeatureDescriptor
+        list and we thus not alter these FeatureDescriptor objects themselves.
+
+        """
+        # dict in which we store all the { output_col_name : (UUID, FeatureDescriptor) }
+        # items of our current Featurecollection object
+        feat_col_fd_mapping: Dict[str, Tuple[str, FeatureDescriptor]] = {}
+        for (s_names, window, stride), fds in self._feature_desc_dict.items():
+            fd: FeatureDescriptor
+            for fd in fds:
+                # As a single FeatureDescriptor can have multiple output col names, we
+                # create a unique identifier for each FeatureDescriptor (on which we
+                # will apply set-like operations later on to only retain all the unique
+                # FeatureDescriptors)
+                uuid_str = str(uuid.uuid4())
+                for output_name in fd.function.output_names:
+                    # Reconstruct the feature column name
+                    feat_col_name = '__'.join([
+                        '|'.join(s_names) if isinstance(s_names, tuple) else s_names,
+                        output_name,
+                        f'w={timedelta_to_str(window)}_s={timedelta_to_str(stride)}'
+                    ])
+                    feat_col_fd_mapping[feat_col_name] = (uuid_str, fd)
+
+        assert all(fc in feat_col_fd_mapping for fc in feat_cols_to_keep)
+
+        # Collect (uuid, FeatureDescriptor) for the feat_cols_to_keep
+        fd_subset: List[Tuple[str, FeatureDescriptor]] = [
+            feat_col_fd_mapping[fc] for fc in feat_cols_to_keep
+        ]
+
+        # Reduce to unique feature descriptor objects (based on uuid) and create a new
+        # FeatureCollection for their deepcopy's.
+        seen_uuids = set()
+        return FeatureCollection(
+            feature_descriptors=[
+                deepcopy(unique_fd) for unique_fd in
+                {
+                    fd for (uuid_str, fd) in fd_subset
+                    if uuid_str not in seen_uuids and not seen_uuids.add(uuid_str)
+                }
+            ]
+        )
+
     def __repr__(self) -> str:
         """Representation string of a FeatureCollection."""
         feature_keys = sorted(set(k[0] for k in self._feature_desc_dict.keys()))
